# -*- coding: utf-8 -*-
"""
AWS Lambda Packager

This script is used to package the lambda function code into a zip file.
It is an alternative to `sam build` and uses poetry to manage dependencies.

"""
from __future__ import annotations

import logging
import os
import platform
import shutil
import subprocess  # nosec B404
from compileall import compile_dir
from datetime import datetime
from pathlib import Path
from py_compile import PycInvalidationMode
from tempfile import TemporaryDirectory
from typing import (
    List,
    Optional,
    Type,
    Union,
)
from zipfile import ZIP_DEFLATED, ZipFile

from .aws_wrangler import fetch_package
from .dep_analyzer import DepAnalyzer
from .pip_analyzer import PipAnalyzer
from .poetry_analyzer import PoetryAnalyzer
from .util import PLATFORMS, PathType

LOG = logging.getLogger(__name__)
MAX_LAMBDA_SIZE = 250 * 1024 * 1024  # 250MB


class LambdaPackager:
    # pylint: disable=too-many-instance-attributes
    layer_dir: PathType | None

    def __init__(
        self,
        project_path: PathType,
        output_dir: PathType,
        python_version: str = "3.9",
        architecture: str = "x86_64",
        region: str = "us-east-1",
        update_dependencies: bool = False,
        ignore_packages: bool = False,
        split_layer: bool = False,
    ):  # pylint: disable=too-many-arguments
        """Initialize the Lambda Packager

        Args:
            project_path: Path to the pyproject.toml file
            python_version: Python version to target
            architecture: Architecture to target (x86_64 or arm64)
            region: AWS region to target
            update_dependencies: whether to update pyproject.toml with the appropriate versions of packages
                from the AWS lambda environment (ignored if ignore_packages is False)
            ignore_packages: Ignore packages that already exist in the AWS lambda environment
        """
        self._reqs = None
        self._pip = None
        self.output_dir = Path(output_dir)
        if ("python" + python_version, architecture) not in PLATFORMS:
            raise Exception(f"{architecture} {python_version} not supported")  # pragma: no cover
        self.project_path = Path(project_path)
        self.python_version = python_version
        self.architecture = architecture
        self.region = region
        self.update_dependencies = update_dependencies
        self.ignore_packages = ignore_packages
        self.split_layer = split_layer
        analyzer_type: Type[DepAnalyzer]
        if (self.project_path / "pyproject.toml").exists() and not (self.project_path / "requirements.txt").exists():
            LOG.info("pyproject.toml found and not requirements.txt, assuming poetry")
            analyzer_type = PoetryAnalyzer
        elif (self.project_path / "requirements.txt").exists() and not (self.project_path / "pyproject.toml").exists():
            LOG.info("requirements.txt found, assuming pip")
            analyzer_type = PipAnalyzer
        else:
            raise Exception("Ambiguous project type, quitting")

        self.analyzer = analyzer_type(
            self.project_path,
            python_version=self.python_version,
            architecture=self.architecture,
            region=self.region,
            ignore_packages=self.ignore_packages,
            update_dependencies=self.update_dependencies,
        )

    @classmethod
    def _get_dir_size(cls, d):
        total_size = 0
        for dirpath, _, filenames in os.walk(d):  # noqa: B007
            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)
        return total_size

    def get_total_size(self):
        total = self._get_dir_size(self.output_dir)
        # if self.layer_dir:
        #     total += self._get_dir_size(self.layer_dir)
        return total

    def get_aws_wrangler_pyarrow(self):
        if "pyarrow" not in self.analyzer.requirements:
            LOG.warning(
                "No pyarrow requirement found in requirements.txt, not bothering to get the aws_wrangler version"
            )
            return
        vers_str = "==" + self.analyzer.requirements["pyarrow"].version
        for p in self.output_dir.glob("pyarrow*"):
            if p.is_dir():
                shutil.rmtree(p, ignore_errors=True)
            else:
                p.unlink()
        try:
            fetch_package(
                "pyarrow",
                self.output_dir,
                vers_str,
                python_version=self.python_version.lstrip("python"),
                arch=self.architecture,
            )
        except ValueError:
            LOG.warning("pyarrow version %s not found", vers_str)

    def strip_tests(self):
        LOG.warning("Stripping tests")
        for p in self.output_dir.glob("**/*"):
            if p.is_file() and "tests" in p.relative_to(self.output_dir).parts:
                LOG.debug("Stripping test file %s", p)
                p.unlink(missing_ok=True)

    def compile_python(self):
        if self.python_version.lstrip("python") == ".".join(platform.python_version_tuple()[:2]):
            LOG.warning("Compiling package")
            LOG.debug('Target Python version: "%s"', self.python_version)
            LOG.debug('Build Python Version: "%s"', ".".join(platform.python_version_tuple()))
            compile_dir(
                str(self.output_dir.absolute()),
                ddir="",
                quiet=2,
                optimize=2,
                workers=1,
                legacy=True,
                force=True,
                invalidation_mode=PycInvalidationMode.UNCHECKED_HASH,
            )
            return True
        LOG.warning("Not compiling package, python version mismatch")
        return False

    def strip_python(self):
        LOG.warning("Stripping python scripts")
        for p in self.output_dir.glob("**/*"):
            if p.is_file() and p.name.endswith(".py"):
                LOG.debug("Stripping python file %s", p)
                p.unlink()

    def strip_other_files(self):
        LOG.warning("Stripping other files")
        for p in self.output_dir.glob("**/*"):
            if p.is_file() and p.suffix in (".pyx", ".pyi", ".pxi", ".pxd", ".c", ".h", ".cc"):
                LOG.debug("Stripping file %s", p)
                p.unlink()

    def strip_libraries(self):
        try:
            LOG.warning("Stripping libraries")
            strip_command = get_strip_binary(self.architecture)
            for p in self.output_dir.glob("**/*.so*"):
                LOG.debug('Stripping library "%s"', p)
                subprocess.run([strip_command, str(p)])  # nosec: B603 pylint: disable=subprocess-run-check
        except Exception:  # pylint: disable=broad-except
            LOG.error("Failed to strip libraries, perhaps we don't have the 'strip' command?")

    def zip_output(self, zip_output):
        if isinstance(zip_output, bool):
            zip_path = Path(str(self.output_dir) + ".zip")
        else:
            zip_path = Path(zip_output)
        with ZipFile(zip_path, "w", compression=ZIP_DEFLATED, compresslevel=9) as zip_file:
            for f in self.output_dir.glob("**/*"):
                if f.is_file():
                    zip_file.write(f, f.relative_to(self.output_dir))

    def package(  # noqa: C901
        self,
        no_clobber: bool = False,
        zip_output: Union[bool, str] = False,
        compile_python: bool = False,
        use_wrangler_pyarrow: bool = False,
        strip_tests: bool = False,  # pylint: disable=unused-argument
        strip_libraries: bool = False,  # pylint: disable=unused-argument
        strip_python: bool = False,
        strip_other_files: bool = False,  # pylint: disable=unused-argument
    ):  # pylint: disable=too-many-arguments,too-many-branches
        if not no_clobber and os.path.exists(self.output_dir):
            LOG.warning("Output directory %s already exists, removing it", self.output_dir)
            shutil.rmtree(self.output_dir, ignore_errors=True)

        self.analyzer.install_dependencies()

        layer_paths = self.analyzer.get_layer_files()
        self.analyzer.install_root()
        self.analyzer.copy_from_target(self.output_dir)
        initial_size = self.get_total_size()
        LOG.info("Pre-strip size: %s", sizeof_fmt(initial_size))

        if use_wrangler_pyarrow:
            self.get_aws_wrangler_pyarrow()
            new_size = self.get_total_size()
            LOG.info("Switched PyArrow size: %s (%0.1f%%)", sizeof_fmt(new_size), new_size / initial_size * 100)

        if strip_python and not compile_python:
            LOG.warning("Not stripping python, since compile_python is set to False")
            strip_python = False

        self.set_utime()
        if compile_python:
            compiled = self.compile_python()
            if strip_python and not compiled:
                strip_python = False
                LOG.warning("Unable to compile python, not stripping python")
            new_size = self.get_total_size()
            LOG.info("Compiled size: %s (%0.1f%%)", sizeof_fmt(new_size), new_size / initial_size * 100)
        for strip_func in ("strip_python", "strip_tests", "strip_libraries", "strip_other_files"):
            if locals()[strip_func]:
                getattr(self, strip_func)()
                new_size = self.get_total_size()
                LOG.info(
                    "%s done, new size: %s (%0.1f%%)", strip_func, sizeof_fmt(new_size), new_size / initial_size * 100
                )
        if self.split_layer:
            self._layer_splitter(layer_paths)
        size_out = self.get_total_size()
        if size_out > MAX_LAMBDA_SIZE:
            LOG.error(
                "Package size %s exceeds maximum lambda size %s", sizeof_fmt(size_out), sizeof_fmt(MAX_LAMBDA_SIZE)
            )
        else:
            LOG.warning("Package size: %s (%0.1f%%)", sizeof_fmt(size_out), size_out / initial_size * 100)
        if zip_output:
            LOG.warning("Zipping output")
            self.zip_output(zip_output)
        if self.split_layer:
            return self.output_dir / "main", self.output_dir / "layer"
        return self.output_dir, None

    def _layer_splitter(self, layer_paths: List[Path]):
        with TemporaryDirectory() as layer_td, TemporaryDirectory() as main_td:
            for lp in layer_paths:
                lp = self.output_dir / lp
                if not lp.exists():
                    continue
                shutil.move(str(lp.resolve()), layer_td)
            for p in self.output_dir.iterdir():
                shutil.move(str(p.resolve()), main_td)
            main_dir = self.output_dir / "main"
            layer_dir = self.output_dir / "layer"
            shutil.move(layer_td, layer_dir)
            shutil.move(main_td, main_dir)

    def set_utime(self, set_time: Optional[int] = None):
        if set_time is None:
            set_time = int(datetime(2020, 1, 1, 1, 1).timestamp()) * int(1e9)
        for dirpath, _, filenames in os.walk(self.output_dir):  # noqa: B007
            for f in filenames:
                fp = os.path.join(dirpath, f)
                os.utime(fp, ns=(set_time, set_time))


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def get_strip_binary(architecture="x86_64"):
    if architecture == "x86_64":
        c = shutil.which("x86_64-linux-gnu-strip") or shutil.which("strip")

    elif architecture == "arm64":
        c = shutil.which("aarch64-linux-gnu-strip")
    else:
        raise ValueError(f"Unknown architecture {architecture}")

    if c is None:
        arch = "aarch64" if architecture == "arm64" else "x86_64"
        LOG.error(
            'Could not find "strip" binary for architecture "%s", perhaps install it with "apt-get install binutils-%s-linux-gnu"?',
            architecture,
            arch,
        )
        raise FileNotFoundError(f"Could not find strip binary for {architecture}")
    return c



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import os
import re
import sys
from contextlib import contextmanager
from os import PathLike
from typing import Union

import requests

LOG = logging.getLogger(__name__)

_lambda_runtime_regex = re.compile(r"^.*?\|\s*`(python3\.\d+)`\s*\|.*?\|\s*([armx864 \\_,]+)\s*\|.*$")

LAMBDA_RUNTIME_DOCS_URL = (
    "https://raw.githubusercontent.com/awsdocs/aws-lambda-developer-guide/main/doc_source/lambda-runtimes.md"
)

PACKAGE_URL = "https://raw.githubusercontent.com/mumblepins/aws-get-lambda-python-pkg-versions/main/{region}-{python_version}-{architecture}.json"

PathType = Union[str, PathLike]


class ArchitectureUnsupported(Exception):
    """Exception raised when the architecture is not supported"""


def get_lambda_runtimes():
    """Gets a list of supported Lambda runtimes from the AWS docs

    Returns:

    """
    r = requests.get(LAMBDA_RUNTIME_DOCS_URL, timeout=10)
    r.raise_for_status()
    runtimes = []
    for line in r.text.splitlines():
        m = _lambda_runtime_regex.match(line)
        if m:
            runtime = m.group(1)
            archs = m.group(2).split(",")
            archs = [a.strip().replace(r"\_", "_") for a in archs]
            for arch in archs:
                runtimes.append((runtime, arch))
    return runtimes


def check_architecture(architecture):
    """Checks if the given architecture is supported

    Args:
        architecture: Architecture to check

    Returns:
        True if the architecture is supported, False otherwise

    """
    architecture = architecture.lower().strip()
    if architecture == "aarch64":
        architecture = "arm64"
    for _, arch in PLATFORMS:
        if arch == architecture:
            return architecture
    raise ArchitectureUnsupported(f"{architecture} not supported")  # pragma: no cover


def get_python_runtime(architecture="x86_64", target_version=None):
    """Gets an allowed python runtime for the given architecture

    Args:
        architecture: Architecture to target (x86_64 or arm64)
        target_version: Python version to target

    Returns:
        A tuple of (python_version, architecture)

    """
    architecture = check_architecture(architecture)
    if target_version is None:
        py_version = sys.version_info[0:2]
    else:
        if isinstance(target_version, str):
            target_version = target_version.lower().strip("python").strip().split(".")
        py_version = tuple(int(v) for v in target_version)

    filtered_platforms = filter(
        lambda x: x[1] == architecture,
        [(tuple(int(v) for v in pv.strip("python").split(".")), ar) for pv, ar in PLATFORMS],
    )
    allowed_py_version = [a for a, _ in filtered_platforms]
    py_version_constrained = max(min(py_version, max(allowed_py_version)), min(allowed_py_version))
    return py_version_constrained, architecture


@contextmanager
def chdir_cm(path: PathType):
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


@contextmanager
def chgenv_cm(**kwargs):
    old_env = os.environ.copy()
    os.environ.update({k: v for k, v in kwargs.items() if v is not None})
    for k, v in kwargs.items():
        if v is None:
            if k in os.environ:
                del os.environ[k]

    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_env)


PLATFORMS = get_lambda_runtimes()

__all__ = ["PathType", "PLATFORMS", "chdir_cm", "chgenv_cm"]



# -*- coding: utf-8 -*-
import logging
import re
import shutil
import tempfile
from pathlib import Path
from typing import Optional, Union
from zipfile import ZipFile

import fsspec
import requests
from appdirs import user_cache_dir
from packaging.specifiers import SpecifierSet
from packaging.version import InvalidVersion, Version

AWS_WRANGLER_RELEASES = "https://api.github.com/repos/awslabs/aws-data-wrangler/releases"
AWS_LAYER_VERSION_URL = "https://github.com/awslabs/aws-data-wrangler/releases/download/{version}/awswrangler-layer-{version}-py{python_version}{arch}.zip"
LOG = logging.getLogger(__name__)
fs = fsspec.filesystem("blockcache", target_protocol="http", cache_storage=user_cache_dir("lambda-packager"))


def to_version(version):
    try:
        return Version(version)
    except InvalidVersion:
        return None


def get_all_versions(python_version="3.9", arch="x86_64"):
    r = requests.get(AWS_WRANGLER_RELEASES, timeout=10)
    rj = r.json()
    rj = filter(lambda x: to_version(x["tag_name"]) is not None, rj)

    tags = {v["tag_name"]: v for v in sorted(rj, key=lambda x: Version(x["tag_name"]), reverse=True)}

    if arch == "x86_64":
        arch = ""
    elif arch == "arm64":
        arch = "-arm64"
    else:
        raise ValueError("arch must be either 'x86_64' or 'arm64'")
    for v in tags.values():
        for a in v["assets"]:
            if re.match(rf"awswrangler-layer-[\d.]*-py{python_version}{arch}.zip", a["name"]):
                with fs.open(a["browser_download_url"], "rb") as f:
                    with ZipFile(f) as z:
                        z.filename = a["name"]
                        yield z


def fetch_package(
    package_name,
    output_dir: Union[str, Path],
    package_version: Optional[str] = None,
    python_version="3.9",
    arch="x86_64",
):
    if package_version is not None:
        pkg_spec = SpecifierSet(package_version)
    else:
        pkg_spec = None
    for zfh in get_all_versions(python_version, arch):
        file_list = [a for a in zfh.namelist() if a.startswith(f"python/{package_name}")]
        wr_pkg_version = [
            re.sub(rf"^.*{package_name}-(.*?)\.dist-info.*$", r"\1", a) for a in file_list if "dist-info" in a
        ][0]
        if pkg_spec is None or pkg_spec.contains(wr_pkg_version):
            LOG.info("Found package %s-%s in %s", package_name, wr_pkg_version, zfh.filename)
            with tempfile.TemporaryDirectory() as tmpdir:
                zfh.extractall(tmpdir, file_list)
                Path(output_dir).mkdir(parents=True, exist_ok=True)
                for p in (Path(tmpdir) / "python").glob("*"):
                    shutil.copytree(p, Path(output_dir) / p.name, dirs_exist_ok=True)
                return wr_pkg_version
    raise ValueError(f"Could not find package {package_name} with version {package_version}")


if __name__ == "__main__":  # pragma: no cover
    logging.basicConfig(level=logging.INFO)
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("package_name", help="Name of the package to fetch")
    parser.add_argument("output_dir", help="Directory to output the package to")
    parser.add_argument("--package-version", help="Version of aws-data-wrangler to use")
    parser.add_argument("--python-version", help="Python version to use", default="3.9")
    parser.add_argument("--arch", help="Architecture to use", default="x86_64")
    args = parser.parse_args()
    fetch_package(args.package_name, Path(args.output_dir), args.package_version, args.python_version, args.arch)



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import re
import shlex
import shutil
import subprocess  # nosec
import tempfile
from abc import ABC, abstractmethod
from collections import namedtuple
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import requests

from .util import PathType, chdir_cm

PackageInfo = namedtuple("PackageInfo", ["name", "version", "version_spec"])
PACKAGE_URL = "https://raw.githubusercontent.com/mumblepins/aws-get-lambda-python-pkg-versions/main/{region}-python{python_version}-{architecture}.json"


class CommandNotFoundError(Exception):
    pass


class ExtraLine(list):
    pass


class DepAnalyzer(ABC):  # pylint: disable=too-many-instance-attributes

    project_root: Path

    analyzer_name: str

    # region init and teardown
    def __init__(
        self,
        project_root: PathType | None,
        python_version: str = "3.9",
        architecture: str = "x86_64",
        region: str = "us-east-1",
        ignore_packages=False,
        update_dependencies=False,
    ):
        self._extra_lines: Optional[List[ExtraLine]] = None
        self._exported_reqs = None
        self._reqs: Optional[Dict[Any, PackageInfo]] = None
        self._pkgs_to_ignore_dict = None
        if project_root is None:
            self.project_root = Path.cwd()
        else:
            self.project_root = Path(project_root)

        self._pip = shutil.which("pip")
        if self._pip is None:
            raise CommandNotFoundError("pip not found, please install and add to PATH")
        self.python_version = python_version
        self.architecture = architecture
        self.region = region

        self.ignore_packages = ignore_packages
        self.update_dependencies = update_dependencies
        self._temp_proj_dir = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        self._chdir = partial(chdir_cm, self._temp_proj_dir.name)
        self._target = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with

        self.log = logging.getLogger(self.__class__.__name__)

    def __del__(self):
        try:
            self._target.cleanup()
            self._temp_proj_dir.cleanup()
        except OSError:
            pass

    # endregion

    # region abstract methods
    @abstractmethod
    def _get_requirements(self) -> Iterable[Union[PackageInfo, ExtraLine]]:
        pass

    @abstractmethod
    def _update_dependency_file(self, pkgs_to_add: Dict[str, PackageInfo]):
        pass

    @abstractmethod
    def direct_dependencies(self) -> Dict[str, str]:
        pass

    # endregion

    # region properties
    @property
    def pkgs_to_ignore_dict(self):
        if not self.ignore_packages:
            return {}
        if self._pkgs_to_ignore_dict is None:
            self._pkgs_to_ignore_dict = self._get_packages_to_ignore()
        return self._pkgs_to_ignore_dict

    @property
    def pkgs_to_ignore(self):
        return [f"{k}=={v}" for k, v in self.pkgs_to_ignore_dict.items()]

    @property
    def pkgs_to_ignore_info(self):
        return {k: PackageInfo(k, v, f"{k}=={v}") for k, v in self.pkgs_to_ignore_dict.items()}

    @property
    def requirements(self) -> Dict[str, PackageInfo]:
        if self._reqs is None:
            self.log.warning("Exporting requirements")
            reqs = self.update_dependency_file()
            if reqs is None:
                reqs = list(self.get_requirements())
            if self._extra_lines is None:
                self._extra_lines = [r for r in reqs if isinstance(r, ExtraLine)]
            self._reqs = {r.name: r for r in reqs if not isinstance(r, ExtraLine)}
        return self._reqs

    @property
    def extra_lines(self):
        if self._extra_lines is None:
            _ = self.requirements
        return self._extra_lines

    # endregion

    # region private methods
    @contextmanager
    def _change_context(self):
        with self._chdir():
            yield

    def _log_popen_output(self, output, level=logging.DEBUG, prefix=""):
        data = ""
        for line in output:
            o = line.decode("utf-8")
            self.log.log(level, prefix + o.rstrip())
            data += o
        return data

    def _install_pip(self, *args, return_state=False, quiet=False, requirements_file=False):
        pip_command = [
            "install",
            "--disable-pip-version-check",
            "--ignore-installed",
            "--no-compile",
            "--python-version",
            self.python_version,
            "--implementation",
            "cp",
        ]
        if not requirements_file:
            for el in self.extra_lines:
                pip_command.extend(el)
        if self.architecture == "arm64":
            pip_command.extend(
                [
                    "--platform",
                    "manylinux2014_aarch64",
                ]
            )
        elif self.architecture == "x86_64":
            pip_command.extend(
                [
                    "--platform",
                    "manylinux2014_x86_64",
                ]
            )
        pip_command.extend(args)
        return self.run_pip(*pip_command, return_state=return_state, quiet=quiet)

    def _get_packages_to_ignore(self):
        try:
            r = requests.get(
                PACKAGE_URL.format(
                    region=self.region, architecture=self.architecture, python_version=self.python_version
                ),
                timeout=30,
            )
            r.raise_for_status()
            data = r.json()
            pkgs_to_ignore_dict = data
        except Exception as e:  # pylint: disable=broad-except
            self.log.warning("Failed to get packages to ignore: %s", e, exc_info=True)
            pkgs_to_ignore_dict = {}
        return pkgs_to_ignore_dict

    # endregion

    # region public methods
    def get_requirements(self) -> Iterable[Union[PackageInfo, ExtraLine]]:
        self.log.info("Getting requirements info using %s", self.analyzer_name)
        return self._get_requirements()

    @classmethod
    def process_requirements(cls, requirements: Iterable[str]) -> Iterable[Union[PackageInfo, ExtraLine]]:
        for line in requirements:
            if line.startswith("#") or line.strip() == "":
                continue
            if line.startswith("-"):
                yield ExtraLine(shlex.split(line))
                continue
            pkg_match = re.match(r"^([^= \n]*)(==)?([^\s;]*).*$", line)
            if pkg_match:
                pkg_name, _, pkg_version = pkg_match.groups()

                yield PackageInfo(pkg_name, pkg_version, line.rstrip())

    def run_command(
        self, *args, return_state=False, quiet=False, prefix=None, context=None
    ) -> Union[bool, Tuple[str, str]]:
        if prefix is None:
            prefix = Path(args[0]).name
        self.log.debug("Running command: %s", args)
        if context is None:
            context = self._change_context
        if quiet:
            loglevel = logging.DEBUG
        else:
            loglevel = logging.INFO
        with context():
            with subprocess.Popen(  # nosec
                [str(a) for a in args],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ) as proc:
                with ThreadPoolExecutor(2) as pool:
                    stdout_fut = pool.submit(
                        self._log_popen_output,
                        proc.stdout,
                        loglevel,
                        prefix + "(OUT)>  ",
                    )
                    stderr_fut = pool.submit(
                        self._log_popen_output,
                        proc.stderr,
                        loglevel,
                        prefix + "(ERR)>  ",
                    )
                    stdout, stderr = stdout_fut.result(), stderr_fut.result()
                proc.wait()
                if return_state:
                    return not bool(proc.returncode)
                if proc.returncode:
                    self.log.error("ERROR IN CALL: %s", args)
                    self.log.error("STDOUT: %s", stdout)
                    self.log.error("STDERR: %s", stderr)
                    raise subprocess.CalledProcessError(proc.returncode, args, stdout, stderr)

                return stdout, stderr

    def update_dependency_file(self):
        if not self.ignore_packages or not self.update_dependencies:
            return None
        self.log.info(
            "Checking to see if any dependencies need to be changed in the dependency file to match the AWS Lambda environment"
        )
        cur_requires = list(self.get_requirements())
        pkgs_to_add = {}
        for pkg in cur_requires:
            if isinstance(pkg, ExtraLine):
                continue
            pkg_name, pkg_version, _ = pkg
            if pkg_name in self.pkgs_to_ignore_dict and pkg_version != self.pkgs_to_ignore_dict[pkg_name]:
                self.log.warning(
                    "%s is currently %s but should be %s", pkg_name, pkg_version, self.pkgs_to_ignore_dict[pkg_name]
                )
                pkgs_to_add[pkg_name] = self.pkgs_to_ignore_info[pkg_name]
        if len(pkgs_to_add) > 0:
            self.log.info("Updating dependency file to add %s requirements", len(pkgs_to_add))
            self._update_dependency_file(pkgs_to_add)
            return None
        self.log.info("No changes needed in the dependency file")
        return cur_requires

    def export_requirements(self):
        if self._exported_reqs is None:
            output = []
            for pkg_name, pkg_version, pkg_spec in self.requirements.values():
                if self.ignore_packages and self.pkgs_to_ignore_dict.get(pkg_name, None) == pkg_version:
                    self.log.warning(
                        "Ignoring %s as it should be in the AWS Lambda Environment already",
                        pkg_spec.strip().split(";")[0],
                    )
                    continue
                output.append(pkg_spec)
            self._exported_reqs = output
        return self._exported_reqs

    def run_pip(self, *args, return_state=False, quiet=False, context=None):
        self.run_command(self._pip, *args, return_state=return_state, quiet=quiet, context=context)

    def install_dependencies(self, quiet=True):
        pip_command = [
            "--target",
            self._target.name,
            "--no-deps",
        ]
        pip_command.extend(self.export_requirements())
        self.log.warning("Installing dependencies using pip")
        self._install_pip(*pip_command, quiet=quiet)
        self.log.warning("Installing dependencies done")

    def install_root(self):
        src_path = self.project_root / "src"
        if src_path.exists():
            if (src_path / "__init__.py").exists():
                self.log.warning("src/__init__.py exists, installing as package in target")
                shutil.copytree(src_path, Path(self._target.name) / "src")
            else:
                self.log.warning("src/__init__.py does not exist, installing files from src directly into target")
                shutil.copytree(src_path, Path(self._target.name), dirs_exist_ok=True)
        elif next(self.project_root.glob("*.py"), None):
            for f in self.project_root.glob("*.py"):
                self.log.warning("Copying %s to target", f)
                shutil.copy(f, self._target.name)
        else:
            self.log.warning("No src/__init__.py or *.py files found, no root program is being installed")

    def get_layer_files(self):
        target_path = Path(self._target.name)
        return [a.relative_to(target_path) for a in target_path.iterdir()]

    def copy_from_target(self, dst: PathType):
        self.log.warning("Copying %s from target to %s", self._target.name, dst)
        shutil.copytree(self._target.name, dst)

    def copy_from_temp_dir(self, files: Iterable[str]):
        for f in files:
            fp = Path(self._temp_proj_dir.name) / f
            if fp.exists():
                shutil.copy(fp, self.project_root)

    def copy_to_temp_dir(self, files: Iterable[str]):
        for f in files:
            fp = Path(self.project_root) / f
            if fp.exists():
                shutil.copy(fp, self._temp_proj_dir.name)

    def backup_files(self, files: Iterable[str]):
        date_str = datetime.utcnow().strftime("%Y-%m-%dT%H-%M-%SZ")
        for f in files:
            fp = Path(self.project_root) / f
            if fp.exists():
                shutil.copy(fp, fp.with_suffix(f".{date_str}{fp.suffix}"))

    # endregion



# -*- coding: utf-8 -*-
import argparse
import logging
import sys
from pathlib import Path

from . import __version__
from ._log_formatter import _CustomLogFormatter
from .lambda_packager import LambdaPackager

LOG = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description="AWS Lambda Python Packager", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("project_path", help="Path to project directory with pyproject.toml or requirements.txt")
    parser.add_argument("output_path", help="Path to output directory")

    parser.add_argument("--python-version", "-pyv", help="Python version to target", default="3.9")
    parser.add_argument(
        "--architecture", "-a", help="Architecture to target", default="x86_64", choices=["x86_64", "arm64"]
    )
    parser.add_argument("--region", help="AWS region to target", default="us-east-1")
    parser.add_argument(
        "--verbose", "-v", help="Verbose output (may be specified multiple times)", action="count", default=0
    )
    parser.add_argument(
        "--zip-output", "-z", help="Output zip file in addition to directory", const=True, nargs="?", default=False
    )

    parser.add_argument("--version", "-V", action="version", version=f"%(prog)s {__version__}")

    opt_group = parser.add_argument_group("Optimization Options")
    opt_group.add_argument(
        "--ignore-packages",
        help="Ignore packages that are already present in the AWS Lambda Python runtime",
        action="store_true",
    )
    opt_group.add_argument(
        "--update-dependencies",
        help="Update project dependency file with the ignored packages (ignored if not --ignore-packages)",
        action="store_true",
    )
    opt_group.add_argument(
        "--compile-python",
        help="Compile the python bytecode ",
        action="store_const",
        const=True,
    )
    opt_group.add_argument(
        "--use-aws-pyarrow",
        help="Use AWS wrangler pyarrow (may result in smaller file size). "
        "Pulls from https://github.com/awslabs/aws-data-wrangler/releases/ until it finds a "
        "Lambda layer that includes the proper PyArrow version.",
        action="store_true",
    )
    opt_group.add_argument(
        "--strip-tests",
        help="Strip tests from the package",
        action="store_const",
        const=True,
        default=argparse.SUPPRESS,
    )
    opt_group.add_argument(
        "--strip-libraries",
        help="Strip debugging symbols from libraries",
        action="store_const",
        const=True,
        default=argparse.SUPPRESS,
    )
    opt_group.add_argument(
        "--strip-python",
        help="Strip python scripts from the package (requires --compile-python) (note, may need to set an ENV variable of PYTHONOPTIMIZE=2)",
        action="store_true",
    )
    opt_group.add_argument(
        "--strip-other",
        help="Strip other files from the package ('.pyx', '.pyi', '.pxi', '.pxd', '.c', '.h', '.cc')",
        action="store_const",
        const=True,
        default=argparse.SUPPRESS,
    )
    opt_group.add_argument(
        "--optimize-all",
        "-O",
        help="Turns on all size optimizations (equivalent to --strip-tests --strip-libraries --ignore-packages --update-dependencies --strip-other). "
        "May be specified multiple times. Second time will also enable --compile-python --strip-python --use-aws-pyarrow",
        action="count",
        default=0,
    )
    return parser


def main_cli():  # pragma: no cover
    main_args(sys.argv[1:])


def main_args(args):
    parser = arg_parser()
    args = parser.parse_args(args)
    if args.strip_python and not args.compile_python:
        parser.error("--strip-python requires --compile-python")

    args_dict = vars(args).copy()
    if args_dict["optimize_all"] >= 1:
        args_dict["strip_tests"] = True
        args_dict["strip_libraries"] = True
        args_dict["ignore_packages"] = True
        args_dict["update_dependencies"] = True
        args_dict["strip_other"] = True
    if args_dict["optimize_all"] >= 2:
        args_dict["strip_python"] = True
        args_dict["compile_python"] = True
        args_dict["use_aws_pyarrow"] = True
    ch = logging.StreamHandler()
    ch.setFormatter(_CustomLogFormatter())

    if args_dict["verbose"] <= 0:  # pragma: no cover
        log_level = logging.WARNING
    elif args_dict["verbose"] == 1:  # pragma: no cover
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG
    ch.setLevel(log_level)
    logging.getLogger().setLevel(log_level)
    logging.getLogger().addHandler(ch)
    logging.getLogger("fsspec").setLevel(logging.INFO)
    LOG.debug("Processed Args: %s", args)

    if "optimize_all" in args_dict:
        del args_dict["optimize_all"]

    main(**args_dict)


def main(
    project_path,
    output_path,
    ignore_packages=False,
    update_dependencies=False,
    python_version="3.9",
    architecture="x86_64",
    region="us-east-1",
    zip_output=False,
    compile_python=False,
    use_aws_pyarrow=False,
    strip_tests=False,
    strip_libraries=False,
    strip_python=False,
    strip_other=False,
    **_,
):  # pylint: disable=too-many-arguments,too-many-locals
    pp_path = Path(project_path)
    if pp_path.is_file() and pp_path.name in ("pyproject.toml", "requirements.txt"):
        pp_path = pp_path.parent

    lp = LambdaPackager(
        project_path=pp_path,
        output_dir=output_path,
        ignore_packages=ignore_packages,
        update_dependencies=update_dependencies,
        python_version=python_version,
        architecture=architecture,
        region=region,
    )
    lp.package(
        zip_output=zip_output,
        compile_python=compile_python,
        use_wrangler_pyarrow=use_aws_pyarrow,
        strip_tests=strip_tests,
        strip_libraries=strip_libraries,
        strip_python=strip_python,
        strip_other_files=strip_other,
    )


if __name__ == "__main__":  # pragma: no cover
    main_cli()



# -*- coding: utf-8 -*-
import logging

from colorama import (
    Fore,
    Style,
    init,
)

init()


# https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
class _CustomLogFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    log_format = "%(asctime)s - %(name)s|(%(filename)s:%(lineno)d) - %(levelname)s - %(message)s"
    # log_cli_format = "%(asctime)s [%(levelname)8s] %(level %(message)s (%(filename)s:%(lineno)s)"

    log_date_format = "%Y-%m-%d %H:%M:%S"

    FORMATS = {
        logging.DEBUG: Fore.CYAN + log_format + Fore.RESET,
        logging.INFO: Fore.BLUE + log_format + Fore.RESET,
        logging.WARNING: Fore.YELLOW + log_format + Fore.RESET,
        logging.ERROR: Fore.RED + log_format + Fore.RESET,
        logging.CRITICAL: Fore.RED + Style.BRIGHT + log_format + Fore.RESET,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt=self.log_date_format)
        return formatter.format(record)



# -*- coding: utf-8 -*-
from __future__ import annotations

import tempfile
from importlib.metadata import distributions
from pathlib import Path
from typing import (
    Dict,
    Iterable,
    Union,
)

from .dep_analyzer import (
    DepAnalyzer,
    ExtraLine,
    PackageInfo,
)
from .util import PathType


def get_packages(path):
    if not isinstance(path, list):
        path = [str(path)]
    dists = distributions(path=path)
    # noinspection PyProtectedMember
    return {b.metadata["Name"]: b.version for b in dists}


class PipAnalyzer(DepAnalyzer):
    analyzer_name = "pip"

    def __init__(
        self,
        project_root: PathType | None,
        python_version: str = "3.9",
        architecture: str = "x86_64",
        region: str = "us-east-1",
        ignore_packages=False,
        update_dependencies=False,
    ):
        super().__init__(project_root, python_version, architecture, region, ignore_packages, update_dependencies)
        # try:
        #     import pkg_resources
        # except ImportError:
        #     self.log.error("pip is not installed")
        #     raise
        self.copy_to_temp_dir(("requirements.txt",))

    def _get_requirements(self) -> Iterable[Union[PackageInfo, ExtraLine]]:
        with tempfile.TemporaryDirectory() as tmpdir:
            self._install_pip(
                "--only-binary=:all:",
                "--target",
                tmpdir,
                "-r",
                Path(self._temp_proj_dir.name) / "requirements.txt",
                quiet=True,
                requirements_file=True,
            )

            for pkg, version in get_packages(tmpdir).items():
                yield PackageInfo(pkg, version, f"{pkg}=={version}")

    @property
    def extra_lines(self):

        self._extra_lines = []
        req_file = self.project_root / "requirements.txt"
        with req_file.open() as f:
            for line in self.process_requirements(f):
                if not isinstance(line, ExtraLine):
                    continue
                self._extra_lines.append(line)
        return self._extra_lines

    def _update_dependency_file(self, pkgs_to_add: Dict[str, PackageInfo]):
        self.backup_files(["requirements.txt"])
        self.log.debug("Updating requirements.txt with %s", ", ".join(f"{k}=={v}" for k, v in pkgs_to_add.items()))
        new_lines = []
        with open(Path(self._temp_proj_dir.name) / "requirements.txt", "r", encoding="utf8") as f:
            for pkg in self.process_requirements(f):
                if isinstance(pkg, ExtraLine):
                    new_lines.append(" ".join(pkg))
                    continue
                if pkg.name not in pkgs_to_add:
                    new_lines.append(pkg.version_spec)
            for pkg in pkgs_to_add.values():
                new_lines.append(pkg.version_spec)
        with open(Path(self._temp_proj_dir.name) / "requirements.txt", "w", encoding="utf8") as f:
            f.write("\n".join(new_lines))
            f.write("\n")
        self.copy_from_temp_dir(["requirements.txt"])

    def direct_dependencies(self) -> Dict[str, str]:
        with (self.project_root / "requirements.txt").open() as f:
            ret = {}
            for pkg in self.process_requirements(f):
                if isinstance(pkg, ExtraLine):
                    continue
                ret[pkg.name] = pkg.version
            return ret



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aws_lambda_python_packager']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp',
 'appdirs',
 'colorama>=0.4',
 'fsspec',
 'packaging',
 'requests',
 'toml>=0.10']

extras_require = \
{':platform_system == "Windows"': ['python-certifi-win32']}

entry_points = \
{'console_scripts': ['lambda-packager = '
                     'aws_lambda_python_packager.__main__:main_cli']}

setup_kwargs = {
    'name': 'aws-lambda-python-packager',
    'version': '1.2.2',
    'description': 'Description',
    'long_description': "# AWS Lambda Python Packager\n\n[![Checks][checks-shield]][checks-url]\n[![Codecov][codecov-shield]][codecov-url]\n\n\n\nAn alternate way to package Python functions for AWS Lambda. Works cross-platform and cross-architecture if binary packages are available for all packages.\n\n```shell\n$ lambda-packager -h\nusage: lambda-packager [-h] [--ignore-packages] [--update-dependencies]\n                       [--python-version PYTHON_VERSION] [--architecture {x86_64,arm64}]\n                       [--region REGION] [--verbose] [--zip-output [ZIP_OUTPUT]] [--version]\n                       [--compile-python] [--use-aws-pyarrow] [--strip-tests] [--strip-libraries]\n                       [--strip-python] [--strip-other] [--optimize-all]\n                       pyproject_path output_path\n\nAWS Lambda Python Packager\n\npositional arguments:\n  pyproject_path        Path to pyproject.toml\n  output_path           Path to output directory\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --ignore-packages     Ignore packages that are already present in the AWS Lambda Python runtime\n                        (default: False)\n  --update-dependencies\n                        Update project dependency file with the ignored packages (ignored if not\n                        --ignore-packages) (default: False)\n  --python-version PYTHON_VERSION, -pyv PYTHON_VERSION\n                        Python version to target (default: 3.9)\n  --architecture {x86_64,arm64}, -a {x86_64,arm64}\n                        Architecture to target (default: x86_64)\n  --region REGION       AWS region to target (default: us-east-1)\n  --verbose, -v         Verbose output (may be specified multiple times) (default: 0)\n  --zip-output [ZIP_OUTPUT], -z [ZIP_OUTPUT]\n                        Output zip file in addition to directory (default: False)\n  --version, -V         show program's version number and exit\n\nOptimization Options:\n  --compile-python      Compile the python bytecode (default: None)\n  --use-aws-pyarrow     Use AWS wrangler pyarrow (may result in smaller file size). Pulls from\n                        https://github.com/awslabs/aws-data-wrangler/releases/ until it finds a\n                        Lambda layer that includes the proper PyArrow version. (default: False)\n  --strip-tests         Strip tests from the package\n  --strip-libraries     Strip debugging symbols from libraries\n  --strip-python        Strip python scripts from the package (requires --compile-python) (note,\n                        may need to set an ENV variable of PYTHONOPTIMIZE=2) (default: False)\n  --strip-other         Strip other files from the package ('.pyx', '.pyi', '.pxi', '.pxd', '.c',\n                        '.h', '.cc')\n  --optimize-all, -O    Turns on all size optimizations (equivalent to --strip-tests --strip-\n                        libraries --ignore-packages --update-pyproject --strip-other). May be\n                        specified multiple times. Second time will also enable --compile-python\n                        --strip-python --use-aws-pyarrow (default: 0)\n\n```\n\n\n\n[codecov-shield]: https://img.shields.io/codecov/c/github/mumblepins/aws-lambda-python-packager\n[codecov-url]: https://app.codecov.io/gh/mumblepins/aws-lambda-python-packager\n\n[checks-shield]: https://img.shields.io/github/workflow/status/mumblepins/aws-lambda-python-packager/Python%20Publish?style=flat-square\n[checks-url]: https://github.com/mumblepins/aws-lambda-python-packager/actions/workflows/python-publish.yml\n",
    'author': 'Daniel Sullivan',
    'author_email': 'mumblepins@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mumblepins/aws-lambda-python-packager/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from __future__ import annotations

import os
import re
import shutil
import tempfile
from contextlib import contextmanager
from typing import (
    Dict,
    Iterable,
    Union,
)

import toml

from .dep_analyzer import (
    CommandNotFoundError,
    DepAnalyzer,
    ExtraLine,
    PackageInfo,
)
from .util import (
    PathType,
    chdir_cm,
    chgenv_cm,
)


class PoetryAnalyzer(DepAnalyzer):
    analyzer_name = "poetry"

    def __del__(self):
        super().__del__()
        self._poetry_env.cleanup()

    def __init__(
        self,
        project_root: PathType | None,
        python_version: str = "3.9",
        architecture: str = "x86_64",
        region: str = "us-east-1",
        ignore_packages=False,
        update_dependencies=False,
    ):
        super().__init__(project_root, python_version, architecture, region, ignore_packages, update_dependencies)
        self._poetry = shutil.which("poetry")
        if self._poetry is None:
            raise CommandNotFoundError("poetry not found, please install and add to PATH")
        self.copy_to_temp_dir(("poetry.lock", "pyproject.toml"))

        self._poetry_env = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        self._env_kwargs = None

    @contextmanager
    def _chgenv(self):
        if self._env_kwargs is None:
            kwargs = {
                "POETRY_VIRTUALENVS_IN_PROJECT": "false",
                "POETRY_VIRTUALENVS_PATH": self._poetry_env.name,
                "VIRTUAL_ENV": None,
            }
            kwargs.update(self._get_credentials())
            self._env_kwargs = kwargs
        with chgenv_cm(**self._env_kwargs):
            yield

    def locked(self):
        return self.run_poetry("lock", "--check", return_state=True, quiet=True)

    def lock(self):
        return self.run_poetry("lock", "--no-update", quiet=True)

    def _get_requirements(self) -> Iterable[Union[PackageInfo, ExtraLine]]:
        output_file = None
        if not self.locked():
            self.log.info("Locking dependencies")
            self.lock()
        try:
            reqs, _ = self.run_poetry("export", "--without-hashes", "--with-credentials", "--only", "main")
            # with self._change_context():
            #     reqs = export_requirements(Path(self._temp_proj_dir.name))
            #
            # self.log.debug(  # pylint: disable=logging-not-lazy
            #     "\n" + "".join([f"requirements.txt>>  {a}" for a in reqs.splitlines(keepends=True)])
            # )

            reqs = [r for r in reqs.splitlines(keepends=True) if not r.startswith("Using p")]
            yield from self.process_requirements(reqs)
        finally:
            if output_file:
                os.remove(output_file.name)

    def _update_dependency_file(self, pkgs_to_add: Dict[str, PackageInfo]):
        self.backup_files(["pyproject.toml", "poetry.lock"])
        self.log.debug("Updating pyproject.toml with %s", ", ".join(f"{k}=={v}" for k, v in pkgs_to_add.items()))
        self.run_poetry("add", "--lock", *[f"{k}=={v.version}" for k, v in pkgs_to_add.items()])
        self.copy_from_temp_dir(["poetry.lock", "pyproject.toml"])

    @contextmanager
    def _change_context(self):
        with self._chdir(), self._chgenv():
            yield

    @contextmanager
    def _change_project_root_context(self):
        with chdir_cm(self.project_root), self._chgenv():
            yield

    def run_poetry(self, *args, return_state=False, quiet=False, context=None):
        return self.run_command(self._poetry, *args, return_state=return_state, quiet=quiet, context=context)

    def install_root(self):
        initial_dist = {(a, a.lstat()) for a in (self.project_root / "dist").glob("*.tar.gz")}
        self.log.debug("Trying to build package with poetry")

        packaged = self.run_poetry(
            "build", "--format", "sdist", quiet=True, return_state=True, context=self._change_project_root_context
        )
        if packaged:
            self.log.info("Package built with poetry, installing")
            final_dist = {(a, a.lstat()) for a in (self.project_root / "dist").glob("*.tar.gz")}
            pkg = next(iter(final_dist - initial_dist))[0].absolute()
            pip_command = ["--target", self._target.name, "--no-deps", pkg]
            self.log.warning("Installing poetry package using pip in target")
            self._install_pip(*pip_command)
            self.log.warning("Installing poetry package done")
        else:
            self.log.warning("Package not built with poetry, falling back to .py files")
            super().install_root()

    def load_toml(self) -> Dict:
        pyproject = self.project_root / "pyproject.toml"
        with pyproject.open() as f:
            data = toml.load(f)
            return data

    def _get_credentials(self) -> Dict[str, str]:
        t = self.load_toml()
        out = {}
        if (
            "tool" in t
            and "aws-deployment" in t["tool"]
            and "source" in t["tool"]["aws-deployment"]
            and isinstance(t["tool"]["aws-deployment"]["source"], dict)
        ):
            for src_name, src_cfg in t["tool"]["aws-deployment"]["source"].items():
                src_name = re.sub(r"[^A-Z0-9]", "_", src_name.upper())

                for k, v in src_cfg.items():
                    k = re.sub(r"[^A-Z0-9]", "_", k.upper())
                    if k == "TOKEN":
                        out[f"POETRY_PYPI_TOKEN_{src_name}"] = v
                    else:
                        out[f"POETRY_HTTP_BASIC_{src_name}_{k}"] = v
        return out

    def direct_dependencies(self) -> Dict[str, str]:
        data = self.load_toml()
        return data["tool"]["poetry"]["dependencies"]
